//==================================
// PEDUMP - Matt Pietrek 1997
// FILE: COMMON.C
//==================================

#include "Common.h"
#include "ExtrnVar.h"

const IMAGE_DEBUG_MISC          *g_pMiscDebugInfo = NULL;
const IMAGE_COFF_SYMBOLS_HEADER *g_pCOFFHeader = NULL;
const DWORD                     *g_pCVHeader = NULL;

COFFSymbolTable * g_pCOFFSymbolTable = NULL;

#if 0
  // Marked as obsolete in MSDN CD 9
  // Bitfield values and names for the LoaderFlags flags
  static const struct DWORD_FLAG_DESCRIPTIONS LoaderFlags[] = {
                    { IMAGE_LOADER_FLAGS_BREAK_ON_LOAD, "BREAK_ON_LOAD" },
                    { IMAGE_LOADER_FLAGS_DEBUG_ON_LOAD, "DEBUG_ON_LOAD" }
                  };

  #define NUMBER_LOADER_FLAGS   DIM(LoaderFlags)
#endif


/*----------------------------------------------------------------------------*/
//
// Header related stuff
//
/*----------------------------------------------------------------------------*/

// Bitfield values and names for the IMAGE_FILE_HEADER flags
static const WORD_FLAG_DESCRIPTIONS ImageFileHeaderCharacteristics[] =
{
{ IMAGE_FILE_RELOCS_STRIPPED, "RELOCS_STRIPPED" },
{ IMAGE_FILE_EXECUTABLE_IMAGE, "EXECUTABLE_IMAGE" },
{ IMAGE_FILE_LINE_NUMS_STRIPPED, "LINE_NUMS_STRIPPED" },
{ IMAGE_FILE_LOCAL_SYMS_STRIPPED, "LOCAL_SYMS_STRIPPED" },
{ IMAGE_FILE_AGGRESIVE_WS_TRIM, "AGGRESIVE_WS_TRIM" },
// { IMAGE_FILE_MINIMAL_OBJECT, "MINIMAL_OBJECT" }, // Removed in NT 3.5
// { IMAGE_FILE_UPDATE_OBJECT, "UPDATE_OBJECT" },   // Removed in NT 3.5
// { IMAGE_FILE_16BIT_MACHINE, "16BIT_MACHINE" },   // Removed in NT 3.5
{ IMAGE_FILE_BYTES_REVERSED_LO, "BYTES_REVERSED_LO" },
{ IMAGE_FILE_32BIT_MACHINE, "32BIT_MACHINE" },
{ IMAGE_FILE_DEBUG_STRIPPED, "DEBUG_STRIPPED" },
// { IMAGE_FILE_PATCH, "PATCH" },
{ IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP, "REMOVABLE_RUN_FROM_SWAP" },
{ IMAGE_FILE_NET_RUN_FROM_SWAP, "NET_RUN_FROM_SWAP" },
{ IMAGE_FILE_SYSTEM, "SYSTEM" },
{ IMAGE_FILE_DLL, "DLL" },
{ IMAGE_FILE_UP_SYSTEM_ONLY, "UP_SYSTEM_ONLY" },
{ IMAGE_FILE_BYTES_REVERSED_HI, "BYTES_REVERSED_HI" }
};

#define NUMBER_IMAGE_HEADER_FLAGS DIM(ImageFileHeaderCharacteristics)

//
// Dump the IMAGE_FILE_HEADER for a PE file or an OBJ
//
void DumpHeader (const IMAGE_FILE_HEADER *pImageFileHeader)
{
    UINT   i, headerFieldWidth = 30;
    time_t ts = *(ULONG*) &pImageFileHeader->TimeDateStamp;
    const char *t_str = "<?>";

    if (ts > 0 && ts < 0xFFFFFFFF)
       t_str =  ctime (&ts);

    printf ("File Header\n");

    printf ("  %-*s%04X (%s)\n", headerFieldWidth, "Machine:",
            pImageFileHeader->Machine,
            GetMachineTypeName(pImageFileHeader->Machine));

    printf ("  %-*s%04X\n", headerFieldWidth, "Number of Sections:",
            pImageFileHeader->NumberOfSections);

    printf ("  %-*s%08lX -> %.24s\n", headerFieldWidth, "TimeDateStamp:",
            pImageFileHeader->TimeDateStamp, t_str);

    printf ("  %-*s%08lX\n", headerFieldWidth, "PointerToSymbolTable:",
            pImageFileHeader->PointerToSymbolTable);

    printf ("  %-*s%08lX\n", headerFieldWidth, "NumberOfSymbols:",
            pImageFileHeader->NumberOfSymbols);

    printf ("  %-*s%04X\n", headerFieldWidth, "SizeOfOptionalHeader:",
            pImageFileHeader->SizeOfOptionalHeader);

    printf ("  %-*s%04X\n", headerFieldWidth, "Characteristics:",
            pImageFileHeader->Characteristics);

    for (i = 0; i < NUMBER_IMAGE_HEADER_FLAGS; i++)
    {
        if (pImageFileHeader->Characteristics & ImageFileHeaderCharacteristics[i].flag)
           printf ("    %s\n", ImageFileHeaderCharacteristics[i].name);
    }
}

const char *GetMachineTypeName (WORD wMachineType)
{
    switch (wMachineType)
    {
        case IMAGE_FILE_MACHINE_I386:
             return "i386";
        case IMAGE_FILE_MACHINE_AMD64:
             return "x64";
        case IMAGE_FILE_MACHINE_IA64:
             return "IA64";
        case IMAGE_FILE_MACHINE_ARM:
             return "ARM";
        case IMAGE_FILE_MACHINE_ARM64:
             return "ARM64";

     // case IMAGE_FILE_MACHINE_I860:
     //      return "i860";
        case IMAGE_FILE_MACHINE_R3000:
             return "R3000";
        case 160:
             return "R3000 big endian";
        case IMAGE_FILE_MACHINE_R4000:
             return "R4000";
        case IMAGE_FILE_MACHINE_R10000:
             return "R10000";
        case IMAGE_FILE_MACHINE_ALPHA:
             return "Alpha";
        case IMAGE_FILE_MACHINE_POWERPC:
             return "PowerPC";
        default:
             return "unknown";
    }
}


/*----------------------------------------------------------------------------*/
//
// Section related stuff
//
/*----------------------------------------------------------------------------*/

//
// These aren't defined in the NT 4.0 WINNT.H, so we'll define them here
//
#ifndef IMAGE_SCN_TYPE_DSECT
#define IMAGE_SCN_TYPE_DSECT                 0x00000001  // Reserved.
#endif

#ifndef IMAGE_SCN_TYPE_NOLOAD
#define IMAGE_SCN_TYPE_NOLOAD                0x00000002  // Reserved.
#endif

#ifndef IMAGE_SCN_TYPE_GROUP
#define IMAGE_SCN_TYPE_GROUP                 0x00000004  // Reserved.
#endif

#ifndef IMAGE_SCN_TYPE_COPY
#define IMAGE_SCN_TYPE_COPY                  0x00000010  // Reserved.
#endif

#ifndef IMAGE_SCN_TYPE_OVER
#define IMAGE_SCN_TYPE_OVER                  0x00000400  // Reserved.
#endif

#ifndef IMAGE_SCN_MEM_PROTECTED
#define IMAGE_SCN_MEM_PROTECTED              0x00004000
#endif

#ifndef IMAGE_SCN_MEM_SYSHEAP
#define IMAGE_SCN_MEM_SYSHEAP                0x00010000
#endif

// Bitfield values and names for the IMAGE_SECTION_HEADER flags
static const struct DWORD_FLAG_DESCRIPTIONS SectionCharacteristics[] = {
                                   { IMAGE_SCN_TYPE_DSECT, "DSECT" },
                                   { IMAGE_SCN_TYPE_NOLOAD, "NOLOAD" },
                                   { IMAGE_SCN_TYPE_GROUP, "GROUP" },
                                   { IMAGE_SCN_TYPE_NO_PAD, "NO_PAD" },
                                   { IMAGE_SCN_TYPE_COPY, "COPY" },
                                   { IMAGE_SCN_CNT_CODE, "CODE" },
                                   { IMAGE_SCN_CNT_INITIALIZED_DATA, "INITIALIZED_DATA" },
                                   { IMAGE_SCN_CNT_UNINITIALIZED_DATA, "UNINITIALIZED_DATA" },
                                   { IMAGE_SCN_LNK_OTHER, "OTHER" },
                                   { IMAGE_SCN_LNK_INFO, "INFO" },
                                   { IMAGE_SCN_TYPE_OVER, "OVER" },
                                   { IMAGE_SCN_LNK_REMOVE, "REMOVE" },
                                   { IMAGE_SCN_LNK_COMDAT, "COMDAT" },
                                   { IMAGE_SCN_MEM_PROTECTED, "PROTECTED" },
                                   { IMAGE_SCN_MEM_FARDATA, "FARDATA" },
                                   { IMAGE_SCN_MEM_SYSHEAP, "SYSHEAP" },
                                   { IMAGE_SCN_MEM_PURGEABLE, "PURGEABLE" },
                                   { IMAGE_SCN_MEM_LOCKED, "LOCKED" },
                                   { IMAGE_SCN_MEM_PRELOAD, "PRELOAD" },
                                   { IMAGE_SCN_LNK_NRELOC_OVFL, "NRELOC_OVFL" },
                                   { IMAGE_SCN_MEM_DISCARDABLE, "DISCARDABLE" },
                                   { IMAGE_SCN_MEM_NOT_CACHED, "NOT_CACHED" },
                                   { IMAGE_SCN_MEM_NOT_PAGED, "NOT_PAGED" },
                                   { IMAGE_SCN_MEM_SHARED, "SHARED" },
                                   { IMAGE_SCN_MEM_EXECUTE, "EXECUTE" },
                                   { IMAGE_SCN_MEM_READ, "READ" },
                                   { IMAGE_SCN_MEM_WRITE, "WRITE" },
                                 };

#define NUMBER_SECTION_CHARACTERISTICS DIM(SectionCharacteristics)

//
// Dump the section table from a PE file or an OBJ
//
void DumpSectionTable (const IMAGE_SECTION_HEADER *section,
                       unsigned cSections, BOOL IsEXE)
{
    unsigned    i, j;
    const char *pszAlign;

    printf ("Section Table\n");

    for (i = 1; i <= cSections; i++, section++)
    {
        printf ("  %02X %-8.8s  %s: %08lX  VirtAddr:  %08lX\n",
                i, section->Name,
                IsEXE ? "VirtSize" : "PhysAddr",
                section->Misc.VirtualSize, section->VirtualAddress);

        printf ("    raw data offs:   %08lX  raw data size: %08lX\n",
                section->PointerToRawData, section->SizeOfRawData);

        printf ("    relocation offs: %08lX  relocations:   %08X\n",
                section->PointerToRelocations, section->NumberOfRelocations);

        printf ("    line # offs:     %08lX  line #'s:      %08X\n",
                section->PointerToLinenumbers, section->NumberOfLinenumbers);

        printf ("    characteristics: %08lX\n", section->Characteristics);

        printf ("    ");
        for (j = 0; j < NUMBER_SECTION_CHARACTERISTICS; j++)
        {
            if (section->Characteristics & SectionCharacteristics[j].flag)
               printf ("  %s", SectionCharacteristics[j].name);
        }

        switch (section->Characteristics & IMAGE_SCN_ALIGN_64BYTES)
        {
            case IMAGE_SCN_ALIGN_1BYTES:
                  pszAlign = "ALIGN_1BYTES";
                  break;
            case IMAGE_SCN_ALIGN_2BYTES:
                 pszAlign = "ALIGN_2BYTES";
                 break;
            case IMAGE_SCN_ALIGN_4BYTES:
                 pszAlign = "ALIGN_4BYTES";
                 break;
            case IMAGE_SCN_ALIGN_8BYTES:
                 pszAlign = "ALIGN_8BYTES";
                 break;
            case IMAGE_SCN_ALIGN_16BYTES:
                 pszAlign = "ALIGN_16BYTES";
                 break;
            case IMAGE_SCN_ALIGN_32BYTES:
                 pszAlign = "ALIGN_32BYTES";
                 break;
            case IMAGE_SCN_ALIGN_64BYTES:
                 pszAlign = "ALIGN_64BYTES";
                 break;
            default:
                 pszAlign = "ALIGN_DEFAULT(16)";
                 break;
        }
        printf ("  %s", pszAlign);

        printf ("\n\n");
    }
}

#ifdef NOT_USED
//
// Given a section name, look it up in the section table and return a
// pointer to the start of its raw data area.
//
LPVOID GetSectionPtr (const char *name, const IMAGE_NT_HEADERS *pNTHeader, DWORD imageBase)
{
    PIMAGE_SECTION_HEADER section = IMAGE_FIRST_SECTION(pNTHeader);
    unsigned i;

    for ( i=0; i < pNTHeader->FileHeader.NumberOfSections; i++, section++ )
    {
        if (strncmp((char *)section->Name, name, IMAGE_SIZEOF_SHORT_NAME) == 0)
            return (LPVOID)(section->PointerToRawData + imageBase);
    }

    return 0;
}
#endif


static const char *SzDebugFormats[] = {
                  "UNKNOWN/BORLAND", "COFF", "CODEVIEW",
                  "FPO", "MISC", "EXCEPTION", "FIXUP",
                  "OMAP_TO_SRC", "OMAP_FROM_SRC"
                };

//
// Dump the debug directory array
//
void DumpDebugDirectory (const IMAGE_DEBUG_DIRECTORY *debugDir, DWORD size, const BYTE *base)
{
    DWORD cDebugFormats = size / sizeof(IMAGE_DEBUG_DIRECTORY);
    const char *szDebugFormat;
    unsigned i;

    if ( cDebugFormats == 0 )
        return;

    printf ("Debug Formats in File\n"
            "  Type            Size     Address  FilePtr  Charactr TimeDate Version\n"
            "  --------------- -------- -------- -------- -------- -------- --------\n");

    for ( i=0; i < cDebugFormats; i++ )
    {
        szDebugFormat = (debugDir->Type <= IMAGE_DEBUG_TYPE_OMAP_FROM_SRC )
                        ? SzDebugFormats[debugDir->Type] : "???";

        printf ("  %-15s %08lX %08lX %08lX %08lX %08lX %u.%02u\n",
                szDebugFormat, debugDir->SizeOfData, debugDir->AddressOfRawData,
                debugDir->PointerToRawData, debugDir->Characteristics,
                debugDir->TimeDateStamp, debugDir->MajorVersion,
                debugDir->MinorVersion);

        switch( debugDir->Type )
        {
           case IMAGE_DEBUG_TYPE_COFF:
                g_pCOFFHeader =
                (const IMAGE_COFF_SYMBOLS_HEADER*)(base+ debugDir->PointerToRawData);
                break;

           case IMAGE_DEBUG_TYPE_MISC:
                g_pMiscDebugInfo =
                (const IMAGE_DEBUG_MISC*)(base + debugDir->PointerToRawData);
                break;

           case IMAGE_DEBUG_TYPE_CODEVIEW:
                g_pCVHeader = (const DWORD*)(base + debugDir->PointerToRawData);
                break;
        }
        debugDir++;
    }
}

/*----------------------------------------------------------------------------*/
//
// Other assorted stuff
//
/*----------------------------------------------------------------------------*/

//
// Do a hexadecimal dump of the raw data for all the sections.  You
// could just dump one section by adjusting the PIMAGE_SECTION_HEADER
// and cSections parameters
//
void DumpRawSectionData (const IMAGE_SECTION_HEADER *section,
                         const void *base, unsigned cSections)
{
    unsigned i;
    char name[IMAGE_SIZEOF_SHORT_NAME + 1];

    printf ("Section Hex Dumps\n");

    for ( i=1; i <= cSections; i++, section++ )
    {
        // Make a copy of the section name so that we can ensure that
        // it's null-terminated
        memcpy(name, section->Name, IMAGE_SIZEOF_SHORT_NAME);
        name[IMAGE_SIZEOF_SHORT_NAME] = 0;

        // Don't dump sections that don't exist in the file!
        if ( section->PointerToRawData == 0 )
            continue;

        printf ("section %02X (%s)  size: %08lX  file offs: %08lX\n",
                i, name, section->SizeOfRawData, section->PointerToRawData);

        HexDump (MakePtr(const BYTE *, base, section->PointerToRawData),
                 section->SizeOfRawData);
        printf ("\n");
    }
}

// Number of hex values displayed per line
#define HEX_DUMP_WIDTH 16

//
// Dump a region of memory in a hexadecimal format
//
void HexDump (const BYTE *ptr, DWORD length)
{
    char buffer[256];
    char *buffPtr, *buffPtr2;
    unsigned cOutput, i;
    DWORD bytesToGo=length;

    while (bytesToGo)
    {
        cOutput = bytesToGo >= HEX_DUMP_WIDTH ? HEX_DUMP_WIDTH : bytesToGo;

        buffPtr = buffer;
        buffPtr += sprintf (buffPtr, "%08lX:  ", length-bytesToGo);
        buffPtr2 = buffPtr + (HEX_DUMP_WIDTH * 3) + 1;

        for ( i=0; i < HEX_DUMP_WIDTH; i++ )
        {
            BYTE value = *(ptr+i);

            if ( i >= cOutput )
            {
                // On last line.  Pad with spaces
                *buffPtr++ = ' ';
                *buffPtr++ = ' ';
                *buffPtr++ = ' ';
            }
            else
            {
                if ( value < 0x10 )
                {
                    *buffPtr++ = '0';
                    itoa( value, buffPtr++, 16);
                }
                else
                {
                    itoa( value, buffPtr, 16);
                    buffPtr+=2;
                }

                *buffPtr++ = ' ';
                *buffPtr2++ = isprint(value) ? value : '.';
            }

            // Put an extra space between the 1st and 2nd half of the bytes
            // on each line.
            if ( i == (HEX_DUMP_WIDTH/2)-1 )
                *buffPtr++ = ' ';
        }

        *buffPtr2 = '\0';  // Null terminate it.
        puts(buffer);      // Can't use printf(), since there may be a '%'
                           // in the string.
        bytesToGo -= cOutput;
        ptr += HEX_DUMP_WIDTH;
    }
}


