#ifndef _COMMON_H
#define _COMMON_H

//==================================
// PEDUMP - Matt Pietrek 1997
// FILE: Common.h
//==================================

#if defined(_MSC_VER)
  #undef  _CRT_NONSTDC_NO_WARNINGS
  #define _CRT_NONSTDC_NO_WARNINGS

  #undef  _CRT_OBSOLETE_NO_WARNINGS
  #define _CRT_OBSOLETE_NO_WARNINGS

  #undef  _CRT_SECURE_NO_WARNINGS
  #define _CRT_SECURE_NO_WARNINGS
#endif

#include <windows.h>
#include <imagehlp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <malloc.h>

#if defined(_MSC_VER) || defined(__MINGW64_VERSION_MAJOR)
  #include <wintrust.h>
  #include <delayimp.h>
#endif


void DumpLibFile       (const void *lpFileBase);
void DumpROMImage      (const IMAGE_ROM_HEADERS *);
void DumpExeFile       (const IMAGE_DOS_HEADER *);
BOOL DumpCVSymbolTable (const BYTE *);
void DumpObjFile       (const IMAGE_FILE_HEADER *);
void DumpDbgFile       (const IMAGE_SEPARATE_DEBUG_HEADER *);
void DumpResources     (const BYTE *pImageBase, const IMAGE_NT_HEADERS *);

class COFFSymbolTable;

void DumpCOFFSymbolTable   (class COFFSymbolTable *pSymTab);
void DumpMiscDebugInfo     (const IMAGE_DEBUG_MISC *MiscDebugInfo);
void DumpCVDebugInfoRecord (const DWORD *pCVHeader);
void DumpLineNumbers       (const IMAGE_LINENUMBER *pln, DWORD count);


/* MakePtr is a macro that allows you to easily add to values (including
 * pointers) together without dealing with C's pointer arithmetic.  It
 * essentially treats the last two parameters as DWORDs.  The first
 * parameter is used to typecast the result to the appropriate pointer type.
 */
#define MakePtr(cast, ptr, addValue) (cast) ((DWORD_PTR)(ptr) + (DWORD_PTR)(addValue))

void DumpHeader (const IMAGE_FILE_HEADER *pImageFileHeader);

void DumpSectionTable (const IMAGE_SECTION_HEADER *section,
                       unsigned cSections, BOOL IsEXE);

//
// Given an RVA, look up the section header that encloses it and return a
// pointer to its IMAGE_SECTION_HEADER
//
template <class T>
const IMAGE_SECTION_HEADER *
      GetEnclosingSectionHeader (DWORD rva, const T *pNTHeader)
{
   const IMAGE_SECTION_HEADER *section = IMAGE_FIRST_SECTION(pNTHeader);
   UINT  i;

   for (i = 0; i < pNTHeader->FileHeader.NumberOfSections; i++, section++)
   {
      // This 3 line idiocy is because Watcom's linker actually sets the
      // Misc.VirtualSize field to 0.  (!!! - Retards....!!!)
      DWORD size = section->Misc.VirtualSize;
      if (size == 0)
         size = section->SizeOfRawData;

      // Is the RVA within this section?
      if ( (rva >= section->VirtualAddress) &&
           (rva < (section->VirtualAddress + size)))
         return (section);
   }
   return (NULL);
}

template <class T>
LPVOID GetPtrFromRVA (DWORD rva, const T *pNTHeader, const BYTE *imageBase)
{
  const IMAGE_SECTION_HEADER *pSectionHdr = GetEnclosingSectionHeader (rva, pNTHeader);

  if (!pSectionHdr)
     return (NULL);

  INT delta = (INT)(pSectionHdr->VirtualAddress - pSectionHdr->PointerToRawData);
  return (PVOID) (imageBase + rva - delta);
}

//
// Like GetPtrFromRVA, but works with addresses that already include the
// default image base
//
template <class T>
LPVOID GetPtrFromVA (const void *ptr, const T* NTHeader, const BYTE *ImageBase)
{
  // Yes, under Win64, we really are lopping off the high 32 bits of a 64 bit
  // value.  We'll knowingly believe that the two pointers are within the
  // same load module, and as such, are RVAs.
  DWORD rva = PtrToLong ((const BYTE*)ptr - NTHeader->OptionalHeader.ImageBase);

  return GetPtrFromRVA (rva, NTHeader, ImageBase);
}


void DumpRawSectionData (const IMAGE_SECTION_HEADER *section,
                         const void *base,
                         unsigned cSections);

void DumpDebugDirectory (const IMAGE_DEBUG_DIRECTORY *debugDir, DWORD size, const BYTE *base);
void DumpCOFFHeader (const IMAGE_COFF_SYMBOLS_HEADER *pDbgInfo);

void HexDump (const BYTE *ptr, DWORD length);

const char *GetMachineTypeName( WORD wMachineType );

#define GetImgDirEntryRVA(pNTHdr, IDE) \
                          (pNTHdr->OptionalHeader.DataDirectory[IDE].VirtualAddress)

#define GetImgDirEntrySize( pNTHdr, IDE ) \
    (pNTHdr->OptionalHeader.DataDirectory[IDE].Size)

extern ULONGLONG      g_mmap_start, g_mmap_end;
extern LARGE_INTEGER  g_file_size;

#define PTR_IN_RANGE(ptr,ofs) \
        ( ( (ULONGLONG)(ptr) + (ofs) < g_mmap_end) && \
          ( (ULONGLONG)(ptr) >= g_mmap_start) )

#define DIM(array)  (sizeof(array) / sizeof(array[0]))

#ifndef IMAGE_SIZEOF_ROM_OPTIONAL_HEADER
#define IMAGE_SIZEOF_ROM_OPTIONAL_HEADER  56
#endif

#ifndef IMAGE_SIZEOF_NT_OPTIONAL64_HEADER
#define IMAGE_SIZEOF_NT_OPTIONAL64_HEADER 240
#endif

#ifndef IMAGE_FILE_MACHINE_ALPHA
#define IMAGE_FILE_MACHINE_ALPHA  0x0184  // Alpha_AXP
#endif

#ifndef IMAGE_FILE_MACHINE_R3000
#define IMAGE_FILE_MACHINE_R3000  0x0162  // MIPS little-endian
#endif

#ifndef IMAGE_FILE_MACHINE_R10000
#define IMAGE_FILE_MACHINE_R10000  0x0168 // MIPS little-endian
#endif

#ifndef IMAGE_DLLCHARACTERISTICS_WDM_DRIVER
#define IMAGE_DLLCHARACTERISTICS_WDM_DRIVER  0x2000     // Driver uses WDM model
#endif

#ifndef IMAGE_REL_BASED_DIR64
#define IMAGE_REL_BASED_DIR64     10
#endif

#if defined(__MINGW32__) && !defined(__MINGW64_VERSION_MAJOR)
  typedef struct  {
          DWORD BeginAddress;
          DWORD EndAddress;
          DWORD UnwindInfoAddress;
        } IMAGE_IA64_RUNTIME_FUNCTION_ENTRY;

  typedef DWORD RVA;

  typedef struct ImgDelayDescr {
          DWORD  grAttrs;        // attributes
          RVA    rvaDLLName;     // RVA to dll name
          RVA    rvaHmod;        // RVA of module handle
          RVA    rvaIAT;         // RVA of the IAT
          RVA    rvaINT;         // RVA of the INT
          RVA    rvaBoundIAT;    // RVA of the optional bound IAT
          RVA    rvaUnloadIAT;   // RVA of optional copy of original IAT
          DWORD  dwTimeStamp;    // 0 if not bound,
        } ImgDelayDescr;

  typedef struct IMAGE_COR20_HEADER
  {
      // Header versioning
      DWORD                   cb;
      WORD                    MajorRuntimeVersion;
      WORD                    MinorRuntimeVersion;

      // Symbol table and startup information
      IMAGE_DATA_DIRECTORY    MetaData;
      DWORD                   Flags;

      // The main program if it is an EXE (not used if a DLL?)
      // If COMIMAGE_FLAGS_NATIVE_ENTRYPOINT is not set, EntryPointToken represents a managed entrypoint.
      // If COMIMAGE_FLAGS_NATIVE_ENTRYPOINT is set, EntryPointRVA represents an RVA to a native entrypoint
      // (depricated for DLLs, use modules constructors intead).
      union {
          DWORD               EntryPointToken;
          DWORD               EntryPointRVA;
      };

      // This is the blob of managed resources. Fetched using code:AssemblyNative.GetResource and
      // code:PEFile.GetResource and accessible from managed code from
      // System.Assembly.GetManifestResourceStream.  The meta data has a table that maps names to offsets into
      // this blob, so logically the blob is a set of resources.
      IMAGE_DATA_DIRECTORY    Resources;
      // IL assemblies can be signed with a public-private key to validate who created it.  The signature goes
      // here if this feature is used.
      IMAGE_DATA_DIRECTORY    StrongNameSignature;

      IMAGE_DATA_DIRECTORY    CodeManagerTable;           // Depricated, not used
      // Used for manged codee that has unmaanaged code inside it (or exports methods as unmanaged entry points)
      IMAGE_DATA_DIRECTORY    VTableFixups;
      IMAGE_DATA_DIRECTORY    ExportAddressTableJumps;

      // null for ordinary IL images.  NGEN images it points at a code:CORCOMPILE_HEADER structure
      IMAGE_DATA_DIRECTORY    ManagedNativeHeader;

  } IMAGE_COR20_HEADER, *PIMAGE_COR20_HEADER;

  #define COMIMAGE_FLAGS_ILONLY             0x00000001
  #define COMIMAGE_FLAGS_32BITREQUIRED      0x00000002
  #define COMIMAGE_FLAGS_IL_LIBRARY         0x00000004
  #define COMIMAGE_FLAGS_STRONGNAMESIGNED   0x00000008
  #define COMIMAGE_FLAGS_NATIVE_ENTRYPOINT  0x00000010
  #define COMIMAGE_FLAGS_TRACKDEBUGDATA     0x00010000
  #define COMIMAGE_FLAGS_ISIBCOPTIMIZED     0x00020000


  #define WIN_CERT_TYPE_X509                  0x0001
  #define WIN_CERT_TYPE_PKCS_SIGNED_DATA      0x0002
  #define WIN_CERT_TYPE_TS_STACK_SIGNED       0x0004

#endif  /* __MINGW32__ && !__MINGW64_VERSION_MAJOR */

typedef DWORD (__stdcall WINAPI *func_UnDecorateSymbolName)
               ( _In_  LPCSTR DecoratedName,
                 _Out_ LPSTR   UnDecoratedName,
                 _In_  DWORD   UndecoratedLength,
                 _In_  DWORD   Flags );

typedef PIMAGE_NT_HEADERS (__stdcall WINAPI *func_CheckSumMappedFile)
                (_In_  void  *BaseAddress,
                 _In_  DWORD  FileLength,
                 _Out_ DWORD *HeaderSum,
                 _Out_ DWORD *CheckSum);

extern func_UnDecorateSymbolName p_UnDecorateSymbolName;
extern func_CheckSumMappedFile   p_CheckSumMappedFile;

const char *undecorate_name (const char *name);

struct WORD_FLAG_DESCRIPTIONS {
       WORD        flag;
       const char *name;
     };

struct DWORD_FLAG_DESCRIPTIONS {
       DWORD       flag;
       const char *name;
     };

#endif  /* _COMMON_H */
