//==================================
// PEDUMP - Matt Pietrek 1997-2001
// FILE: DBGDUMP.CPP
//==================================

#include "Common.h"
#include "COFFSymbolTable.h"
#include "ExtrnVar.h"

static void DumpImageDbgHeader (const IMAGE_SEPARATE_DEBUG_HEADER *pImageSepDbgHeader)
{
    UINT headerFieldWidth = 30;

    printf ("  %-*s%04X\n", headerFieldWidth, "Flags:",
            pImageSepDbgHeader->Flags);

    printf ("  %-*s%04X %s\n", headerFieldWidth, "Machine:",
            pImageSepDbgHeader->Machine,
            GetMachineTypeName(pImageSepDbgHeader->Machine));

    printf ("  %-*s%04X\n", headerFieldWidth, "Characteristics:",
            pImageSepDbgHeader->Characteristics);

    time_t timeStamp = pImageSepDbgHeader->TimeDateStamp;
    const char *t_str = "<none>";

    if (timeStamp > 0 && timeStamp < 0xFFFFFFFF)
       t_str = ctime(&timeStamp);

    printf ("  %-*s%08lX -> %s", headerFieldWidth, "TimeDateStamp:",
            pImageSepDbgHeader->TimeDateStamp, t_str);

    printf ("  %-*s%08lX\n", headerFieldWidth, "CheckSum:",
            pImageSepDbgHeader->CheckSum);

    printf ("  %-*s%08lX\n", headerFieldWidth, "ImageBase:",
            pImageSepDbgHeader->ImageBase);

    printf ("  %-*s%08lX\n", headerFieldWidth, "Size of Image:",
            pImageSepDbgHeader->SizeOfImage);

    printf ("  %-*s%04lX\n", headerFieldWidth, "Number of Sections:",
            pImageSepDbgHeader->NumberOfSections);

    printf ("  %-*s%04lX\n", headerFieldWidth, "ExportedNamesSize:",
            pImageSepDbgHeader->ExportedNamesSize);

    printf ("  %-*s%08lX\n", headerFieldWidth, "DebugDirectorySize:",
            pImageSepDbgHeader->DebugDirectorySize);

    printf ("  %-*s%08lX\n", headerFieldWidth, "SectionAlignment:",
            pImageSepDbgHeader->SectionAlignment);
}

void DumpDbgFile (const IMAGE_SEPARATE_DEBUG_HEADER *pImageSepDbgHeader)
{
    DumpImageDbgHeader (pImageSepDbgHeader);
    printf ("\n");

    DumpSectionTable ((const IMAGE_SECTION_HEADER*) (pImageSepDbgHeader+1),
                      pImageSepDbgHeader->NumberOfSections, TRUE);

    DumpDebugDirectory (
        MakePtr(const IMAGE_DEBUG_DIRECTORY *,
        pImageSepDbgHeader, sizeof(IMAGE_SEPARATE_DEBUG_HEADER) +
        (pImageSepDbgHeader->NumberOfSections * sizeof(IMAGE_SECTION_HEADER)) +
        pImageSepDbgHeader->ExportedNamesSize),
        pImageSepDbgHeader->DebugDirectorySize,
        (const BYTE*)pImageSepDbgHeader);

    printf ("\n");

    if (g_pCOFFHeader)
    {
        DumpCOFFHeader (g_pCOFFHeader);

        printf ("\n");

        g_pCOFFSymbolTable = new COFFSymbolTable (
            MakePtr (PVOID, g_pCOFFHeader, g_pCOFFHeader->LvaToFirstSymbol),
            g_pCOFFHeader->NumberOfSymbols);

        DumpCOFFSymbolTable (g_pCOFFSymbolTable);

        delete g_pCOFFSymbolTable;
    }

    if (g_pCVHeader)
    {
        DumpCVSymbolTable ((const BYTE*)g_pCVHeader);
    }
}
