//==================================
// PEDUMP - Matt Pietrek 1997
// FILE: EXEDUMP.C
//==================================

#include "Common.h"
#include "COFFSymbolTable.h"
#include "ExtrnVar.h"

// VARIATION on the IMAGE_FIRST_SECTION macro from WINNT.H

#define IMAGE_FIRST_ROM_SECTION(ntheader) ( (const IMAGE_SECTION_HEADER*)         \
        ( (DWORD)ntheader +                                                       \
          FIELD_OFFSET(IMAGE_ROM_HEADERS, OptionalHeader) +                       \
          ((const IMAGE_ROM_HEADERS*)(ntheader))->FileHeader.SizeOfOptionalHeader \
        ) )

static void DumpROMOptionalHeader (const IMAGE_ROM_OPTIONAL_HEADER * pROMOptHdr)
{
  UINT width = 30;

  printf ("Optional Header\n");

  printf ("  %-*s%04X\n", width, "Magic", pROMOptHdr->Magic);
  printf ("  %-*s%u.%02u\n", width, "linker version", pROMOptHdr->MajorLinkerVersion, pROMOptHdr->MinorLinkerVersion);
  printf ("  %-*s%lX\n", width, "size of code", pROMOptHdr->SizeOfCode);
  printf ("  %-*s%lX\n", width, "size of initialized data", pROMOptHdr->SizeOfInitializedData);
  printf ("  %-*s%lX\n", width, "size of uninitialized data", pROMOptHdr->SizeOfUninitializedData);
  printf ("  %-*s%lX\n", width, "entrypoint RVA", pROMOptHdr->AddressOfEntryPoint);
  printf ("  %-*s%lX\n", width, "base of code", pROMOptHdr->BaseOfCode);
  printf ("  %-*s%lX\n", width, "base of Bss", pROMOptHdr->BaseOfBss);
  printf ("  %-*s%lX\n", width, "GprMask", pROMOptHdr->GprMask);

  printf ("  %-*s%lX\n", width, "CprMask[0]", pROMOptHdr->CprMask[0]);
  printf ("  %-*s%lX\n", width, "CprMask[1]", pROMOptHdr->CprMask[1]);
  printf ("  %-*s%lX\n", width, "CprMask[2]", pROMOptHdr->CprMask[2]);
  printf ("  %-*s%lX\n", width, "CprMask[3]", pROMOptHdr->CprMask[3]);

  printf ("  %-*s%lX\n", width, "GpValue", pROMOptHdr->GpValue);
}

void DumpROMImage (const IMAGE_ROM_HEADERS * pROMHeader)
{
  DumpHeader (&pROMHeader->FileHeader);
  printf ("\n");

  DumpROMOptionalHeader (&pROMHeader->OptionalHeader);
  printf ("\n");

  DumpSectionTable (IMAGE_FIRST_ROM_SECTION(pROMHeader), pROMHeader->FileHeader.NumberOfSections, TRUE);
  printf ("\n");

  // Dump COFF symbols out here.  Get offsets from the header
}
